\ keyboard.f - Keyboard driver                  22 Jan 2000  tcn


$1100 tCON KBUF

\ Keybuffer format (offset from KBUF):
\  +00 head
\  +04 tail
\  +08 start = KBUF+16
\  +12 end
\  +16 buffer (to end)


\ Misc. functions - move elsewhere:
:' IRQ-ACK ( n -)   \ a macro
  $20 # AL MOV,
  8 >= if  $A0 # AL OUT,  then
  $20 # AL OUT,  ;

TCODE: MAP-INT ( n xt -)
  EBX PUSH,  0 #@ EBP+ EBX MOV,
  3 # EBX SHL,  IDT # EBX ADD,
  AX 0 EBX@ MOV,  16 # EAX SHR,  AX 6 EBX@ MOV,
  tW 2* # EBP ADD, ( 2DROP)  EBX POP,  ;TCODE

TCODE: IRQ-ON ( n -)
  ECX PUSH,  EAX ECX MOV,
  1 # AL MOV,  AL SHL,  AL COM,  AL CL MOV,
  $21 # AL IN,  CL AL AND,  $21 # AL OUT,
  DROP,  ECX POP,  ;TCODE


\ Keymap format:
\   (Designed to be easily modified at runtime)
\ "modifier bitmap --> keymap column" xlat table (256 bytes)
\ Keymap: 128 rows (scancodes)
\      by 256 columns (modifier combinations)
\ Each cell (4 bytes) contains:
\    First 2 bits = Action type:
\       0 = function (cell = aligned address)
\       1 = ASCII code
\       2 = String (cell 2 -  = aligned address)
\       3 = Modifier (cell 8 RSHIFT = bitmask)


\ Scancode definitions:
DECIMAL
1   CONSTANT kESC     2   CONSTANT k1       3   CONSTANT k2
4   CONSTANT k3       5   CONSTANT k4       6   CONSTANT k5
7   CONSTANT k6       8   CONSTANT k7       9   CONSTANT k8
10  CONSTANT k9       11  CONSTANT k0       12  CONSTANT k-
13  CONSTANT k=       14  CONSTANT kBS      15  CONSTANT kTAB
16  CONSTANT kQ       17  CONSTANT kW       18  CONSTANT kE
19  CONSTANT kR       20  CONSTANT kT       21  CONSTANT kY
22  CONSTANT kU       23  CONSTANT kI       24  CONSTANT kO
25  CONSTANT kP       26  CONSTANT k[       27  CONSTANT k]
28  CONSTANT kCR      29  CONSTANT kLCTL    30  CONSTANT kA
31  CONSTANT kS       32  CONSTANT kD       33  CONSTANT kF
34  CONSTANT kG       35  CONSTANT kH       36  CONSTANT kJ
37  CONSTANT kK       38  CONSTANT kL       39  CONSTANT k;
40  CONSTANT k'       41  CONSTANT k`       42  CONSTANT kLSH
\ Scancodes, continued:
43  CONSTANT k\       44  CONSTANT kZ       45  CONSTANT kX
46  CONSTANT kC       47  CONSTANT kV       48  CONSTANT kB
49  CONSTANT kN       50  CONSTANT kM       51  CONSTANT k,
52  CONSTANT k.       53  CONSTANT k/       54  CONSTANT kRSH
55  CONSTANT k*       56  CONSTANT kLALT    57  CONSTANT kSP
58  CONSTANT kCAPS    59  CONSTANT kF1      60  CONSTANT kF2
61  CONSTANT kF3      62  CONSTANT kF4      63  CONSTANT kF5
64  CONSTANT kF6      65  CONSTANT kF7      66  CONSTANT kF8
67  CONSTANT kF9      68  CONSTANT kF10     69  CONSTANT kNUM
70  CONSTANT kSCR     71  CONSTANT kHOME    72  CONSTANT kUP
73  CONSTANT kPGUP    74  CONSTANT k-PAD    75  CONSTANT kLEFT
76  CONSTANT k5PAD    77  CONSTANT kRIGHT   78  CONSTANT k+PAD
79  CONSTANT kEND     80  CONSTANT kDOWN    81  CONSTANT kPGDN
82  CONSTANT kINS     83  CONSTANT kDEL     84  CONSTANT kPRTSC

\ Scancodes for extended keys ("artificial" scancodes)
87  CONSTANT kF11     88  CONSTANT kF12
\ The stupid win95 keys:
91  CONSTANT kLWIN    92  CONSTANT kRWIN    93  CONSTANT kMENU
\ Miscellaneous
100  CONSTANT kPAD
120  CONSTANT kRCTL   121  CONSTANT kRALT
122  CONSTANT k/PAD   123  CONSTANT kPAUSE








\ Modifier definitions:
:' MOD: ( m "name" -) ( m - m')
  CREATE C,
  DOES> C@ OR ;

$01 MOD: mCTL
$02 MOD: mALT
$04 MOD: mSH
$08 MOD: mCAPS
$10 MOD: mNUM






\ ASSIGN COLUMNS:
\ Modifier bitmask --> column number
\       (Default to column 0)
\    8-bit translation table (256 bytes):
\
CREATE COLS   256 ALLOT  COLS 256 ERASE
:' :COL ( column mask - )  COLS + C! ;
:' COL@ ( mask - column )  COLS + C@ ;
  1  0 mCTL                       :COL  \ control chars
  1  0 mCTL      mSH              :COL
  1  0 mCTL          mCAPS        :COL
  1  0 mCTL      mSH mCAPS        :COL
  2  0               mCAPS        :COL  \ uppercase/numbers
  3  0           mSH              :COL  \ uppercase/symbols
  4  0           mSH mCAPS        :COL  \ lowercase/symbols
5 CONSTANT #COLS
\ Copy to target
:' COLS_COPY   COLS 256 0 DO  DUP C@ db, 1+  LOOP DROP ;
LABEL: nCOLS   #COLS dc,
."    tCOLS at " $.
LABEL: tCOLS   COLS_COPY











\ Action types:
:' aCALL ( xt - action)    ( no-op) ;
:' aKEY ( char - action)   8 LSHIFT 1+ ;
\ :' aSTR ( cstr - action)  ... ;
:' aMOD ( m - action)      8 LSHIFT 3 + ;











\ Keymap:
16 0ALIGN,
."    Keymap at " $.
LABEL: KEYMAP
LABEL' KEYMAP'   512 #COLS *  0FILL
:' :KEY ( scancode modifiers action - )
   -ROT DUP -1 =
   IF   DROP 4 * KEYMAP' +  #COLS 0 DO 2DUP dc! 512 + LOOP 2DROP
   ELSE COL@ 128 *  + 4 *  KEYMAP' + dc!
   THEN ;








( scan    mods    action )
(   If mods = -1, all combinations are assigned)
  kLSH    -1      0 mSH   aMOD  :KEY      \ Modifier keys
  kRSH    -1      0 mSH   aMOD  :KEY
  kLCTL   -1      0 mCTL  aMOD  :KEY
  kRCTL   -1      0 mCTL  aMOD  :KEY
  kLALT   -1      0 mALT  aMOD  :KEY
  kRALT   -1      0 mALT  aMOD  :KEY
  kCAPS   -1      0 mCAPS aMOD  :KEY
  kNUM    -1      0 mNUM  aMOD  :KEY






:' :LETTER ( scancode uppercase - )
  2DUP -1             SWAP aKEY :KEY
  $20 XOR ( to lowercase)
  2DUP 0              SWAP aKEY :KEY
  $60 XOR ( to control-char)
       0 mCTL         SWAP aKEY :KEY  ;

kA 'A :LETTER   kB 'B :LETTER   kC 'C :LETTER   kD 'D :LETTER
kE 'E :LETTER   kF 'F :LETTER   kG 'G :LETTER   kH 'H :LETTER
kI 'I :LETTER   kJ 'J :LETTER   kK 'K :LETTER   kL 'L :LETTER
kM 'M :LETTER   kN 'N :LETTER   kO 'O :LETTER   kP 'P :LETTER
kQ 'Q :LETTER   kR 'R :LETTER   kS 'S :LETTER   kT 'T :LETTER
kU 'U :LETTER   kV 'V :LETTER   kW 'W :LETTER   kX 'X :LETTER
kY 'Y :LETTER   kZ 'Z :LETTER

k` 0 '` aKEY :KEY    k` 0 mSH '~ aKEY :KEY \ Top Row
k1 0 '1 aKEY :KEY    k1 0 mSH '! aKEY :KEY
k2 0 '2 aKEY :KEY    k2 0 mSH '@ aKEY :KEY
k3 0 '3 aKEY :KEY    k3 0 mSH '# aKEY :KEY
k4 0 '4 aKEY :KEY    k4 0 mSH '$ aKEY :KEY
k5 0 '5 aKEY :KEY    k5 0 mSH '% aKEY :KEY
k6 0 '6 aKEY :KEY    k6 0 mSH '^ aKEY :KEY
k7 0 '7 aKEY :KEY    k7 0 mSH '& aKEY :KEY
k8 0 '8 aKEY :KEY    k8 0 mSH '* aKEY :KEY
k9 0 '9 aKEY :KEY    k9 0 mSH '( aKEY :KEY
k0 0 '0 aKEY :KEY    k0 0 mSH ') aKEY :KEY
k- 0 '- aKEY :KEY    k- 0 mSH '_ aKEY :KEY
k= 0 '= aKEY :KEY    k= 0 mSH '+ aKEY :KEY

k[ 0 '[ aKEY :KEY    k[ 0 mSH '{ aKEY :KEY \ More punctuation
k] 0 '] aKEY :KEY    k] 0 mSH '} aKEY :KEY
k\ 0 '\ aKEY :KEY    k\ 0 mSH '| aKEY :KEY
k; 0 '; aKEY :KEY    k; 0 mSH ': aKEY :KEY
k' 0 '' aKEY :KEY    k' 0 mSH '" aKEY :KEY
k, 0 ', aKEY :KEY    k, 0 mSH '< aKEY :KEY
k. 0 '. aKEY :KEY    k. 0 mSH '> aKEY :KEY
k/ 0 '/ aKEY :KEY    k/ 0 mSH '? aKEY :KEY

kBS  -1   8 aKEY :KEY
kTAB -1   9 aKEY :KEY
kCR  -1  10 aKEY :KEY
kESC -1  27 aKEY :KEY
kSP  -1  32 aKEY :KEY

\ TBD: arrows, keypad, fn keys..


."    EKEY-TBL at " $.
\ Ext. scancodes --> fake scancodes
\    8-bit translation table (256 bytes):
\
LABEL' EKEY-TBL   256 0FILL
\
:' :EXT ( scan scan' -)
  SWAP EKEY-TBL + db! ;
\
\ TBD: Use mnemonics (ie, kDEL) rather than numbers
\      Also, there are more extended scancodes
28  96 :EXT
29  97 :EXT
53  98 :EXT
56 100 :EXT
71 102 :EXT
72 103 :EXT
73 104 :EXT
75 105 :EXT
77 106 :EXT
79 107 :EXT
80 108 :EXT
81 109 :EXT
82 110 :EXT
83 111 :EXT
111 112 :EXT



\ Keyboard driver:

LABEL: EXT?    0 db,  \ Extended key?
LABEL: PAUSE?  0 db,  \ 
LABEL: MODS    0 db,  \ Modifier bitmask










."    Action-handlers at " $.
\ Actions... TBD:
\ 0. Function -- Call it (or do nothing, if 0)
LABEL: A_CALL
  BL BL OR,  0=, IF,  \ Key down?
\    EAX EAX OR,  0=, NOT, IF,  EAX CALL,  THEN,
  THEN,  RET,

\ 1. ASCII -- add to key buffer
LABEL: A_ASCII
  BL BL OR,  0=, IF,  \ Key down?
    EDI PUSH,
    KEYBUF #@ EDI MOV,  \ HEAD of buffer
    \ TBD: Check for buffer overflow
    ( 8-bit:)  AH AL MOV,          STOS,
\    ( 16-bit:) 8 # EAX SHR,  W-PTR STOS,
    KEYBUF 12 + #@ EAX MOV,
    EAX EDI CMP, 0=, IF, KEYBUF 8 + #@ EDI MOV, THEN, \ Wrap
    EDI KEYBUF #@ MOV,
    EDI POP,
  THEN,  RET,



\ 2. String -- (if implemented)
LABEL: A_STRING
  RET,
\      lookup string at addr; add each char. to key buffer

\ 3. Modifier -- 
LABEL: A_MOD
\ $30 # AH ADD,  AH $B8052 #@ MOV,  $30 # AH SUB,
  BL BL OR,  0=, IF,  AH MODS #@ OR,   RET,  THEN,
                      AH MODS #@ XOR,  RET,

4 0ALIGN,  LABEL: jtable
   A_CALL dc,  A_ASCII dc,  A_STRING dc,  A_MOD dc,



LABEL: KBD-DECODE ( EAX=scancode -)
."    KBD-DECODE at " $.
  EAX EBX MOV,  $7F # AL AND,
  \ Handle extended keys:
\  0 # EXT? #@ CMP,  0=, NOT, IF,
\    \ TBD: convert to "fake" scancode w/ XLAT,
\    0 # EXT? #@ MOV,  \ turn EXT? off
\  THEN,
  \ Determine if pressed/released: (BL=00 if up, 80 if down)
  $80 # BL AND,  BL KEYTBL EAX@ MOV,  \ update keys-down table
  2 # EAX SHL,  \ 4* (get row #)
  \ translate MODS -> column# (tCOLS = addr of xlat table)
  MODS #@ EDX MOVZX,
  tCOLS EDX@ DL MOV,
\ 2 # DL MOV,
  9 # EDX SHL,  \ 128* 4*
  EDX EAX ADD,
  KEYMAP EAX@ EAX MOV,
\ KEYMAP EDX@ EAX+ 4x  EAX  MOV,         \ get keymap cell
  EDX EDX XOR,  AL DL MOV,  3 # DL AND,  \ get action type
  jtable #@ EDX+ 4x JMP,                 \ jump table

\ IRQ HANDLER
LABEL: KBD-IRQ
."    KBD-IRQ at " $.
  EAX PUSH, EBX PUSH, EDX PUSH,
  EAX EAX XOR,  $60 # AL IN,
  0 # PAUSE? #@ CMP,  0=, NOT, if,
    PAUSE? #@ DEC,
  else,  $E1 # AL CMP,  0=, if, ( PAUSE key)
    5 # PAUSE? #@ MOV,
    else,  $E0 # AL CMP,  0=, if, ( extended key)
      -1 # EXT? #@ MOV,
      else,  KBD-DECODE # CALL,  then,
    then,
  then,
  1 IRQ-ACK
  EDX POP, EBX POP, EAX POP,
  IRET,


\ INITIALIZATION
TCODE: KBD-INIT
  EAX PUSH,  ECX PUSH,

  \ Initialize keystroke buffer
  KEYBUF # EAX MOV,  EAX EDI MOV,  16 # EAX ADD,
  D-PTR STOS,  D-PTR STOS,  D-PTR STOS,  \ head, tail, start
  $100 # EAX ADD,  D-PTR STOS,           \ end

  \ Clear key-down table
  EAX EAX XOR,  128 4 / # ECX MOV,  KEYTBL # EDI MOV,
  REP D-PTR STOS,

  ECX POP,  EAX POP,

  \ Enable keyboard IRQ handler
  $21 # DPUSH,  { KBD-IRQ } # DPUSH,  { MAP-INT } # CALL,
  1 # DPUSH,  { IRQ-ON } # CALL,
  ;TCODE
