\ meta.f - Metacompiler                              tcn 02jul00

ONLY FORTH DEFINITIONS
VOCABULARY META   ALSO META DEFINITIONS
VOCABULARY TARGET


: {  \ Enter "compiling to target" mode
 ONLY FORTH  ALSO ASSEMBLER  ALSO META  ALSO TARGET
     DEFINITIONS ;
: }  \ Return to "normal" mode
  ONLY FORTH  ALSO ASSEMBLER  ALSO META
     DEFINITIONS ;
: {'  { ; IMMEDIATE            : }'  } ; IMMEDIATE

\ variable LASTXT   \ xt of last entry added to target dict
\ 0 LASTXT !
2variable lastxt
variable tCONTEXT  \ 0=Main, 1=Compiler

: FORTH    0 tCONTEXT ! ;
: COMPILER 1 tCONTEXT ! ;

variable tSTATE   \ STATE for metacompiler
: [[   } 0 tSTATE !   ;
: ]]     1 tSTATE ! { ;

[[   \ Initial state
FORTH

\ Dictionary headers:
DECIMAL
variable tCFA  \ Current/latest definition's CFA and PFA
variable tPFA
: tCFA! ( - )  $  tCFA @ T>H dc! ;  \ Set CFA to $

: (tHEAD) ( "name" -)  \ Create dict header on target system
  32 WORD COUNT TUCK  ( # a #)          \ Parse name

  $ + 1+ dup aligned - negate           \ Calculate alignment
  ?dup if  0 DO  $90 db,  LOOP  then    \ and pad out...
  OVER 0  DO  DUP C@ db,  CHAR+ LOOP    \ NAME
  DROP db,                              \ NAME LENGTH

  tCONTEXT @ CELLS  lastxt +            \ Link field
  dup @  $ rot !  dc,

  $ tCFA ! ;                            \ Code field
\  $ tPFA !  tCFA! ;                     \ Parameter field

: tHEAD ( "name" -)  \ Create dict header on host & target
  >IN @ { CREATE } >IN !  (tHEAD) ;


\ Target CODE: compiler
: tCODE: ( "name" -) ( -)
  tHEAD
  $ , DOES> @  tSTATE @ if # CALL, then ;
: ;tCODE ( -)       RET, ;
: ;tCODE-IRET ( -)  IRET, ;


\ DOLIT DOCON DOVAR DODOES
tCODE: dolit ( - n)
  0push,
  $8B db, $04 db, $24 db,        \ 0 esp@ eax mov,
  $80 db, $04 db, $24 db, tW db, \ tW # 0 esp@ add,
  0 eax@ eax mov,
  ;tCODE

tCODE: docon ( - n)
  0push,
  eax pop,
  0 eax@ eax mov,
  ;tCODE

tCODE: dovar ( - a)
  0push,
  eax pop,
  ;tCODE

tCODE: dodoes ( R: pfa cfa - cfa) ( - pfa)
  0push,
  ;tCODE

\ tLIT tCON tCREATE tVAL tVAR  (tDOES> is a few screens later)
: tLIT ( n)  [ { dolit } ] LITERAL # call,  dc, ;

: tCON ( n "name") ( - n)
  tHEAD  dup , $ ,  [ { docon } ] LITERAL # call,  dc,
  DOES> tSTATE @  IF   CELL+ @ # call,
                  ELSE @ THEN ;
: tCREATE
  tHEAD  $ , [ { dovar } ] LITERAL # call,
  DOES> @  tSTATE @  IF  # call,  THEN ;

: tVAL ( n "name") ( - a)  tCREATE  dc, ;
: tVAR (   "name") ( - a)  0 tVAL ;

\ Changes a tVAR or tVAL
: t! ( n ta)  T>H dc! ;

\ tCREATE ...
\ (tCON) creates a "constant" on both host & target
\ tVAR creates a regular variable, initialized to 0
\ tVAL creates a preinitialized variable
\ tCON creates a constant



\ METACOMPILER

\ In META vocab, alias : and ; to :' and ;'
\ ;' is probably unnecessary since ; is defined in TARGET
: :'  : ;
: ;'  POSTPONE ;  ; IMMEDIATE

: GET-WORD ( - cstr (Parse next available word)
  BEGIN  32 WORD  DUP C@ 0=
  WHILE  DROP REFILL  DUP 0=
         IF ." Unexpected end of input!" CR ABORT THEN
         DROP
  REPEAT ;

\ STC Forth target compiler (metacompiler, or whatever)
: (:)  ( -) ( -)
  { ]]
  BEGIN
    GET-WORD FIND
    IF ( found) EXECUTE  ( treat everything as immediate)
    ELSE   COUNT 2DUP SNUMBER?
      IF   tSTATE @ if  tLIT then  2DROP
      ELSE TYPE ."  not found in TARGET dict" CR ABORT
    THEN  THEN  tSTATE @  0= UNTIL ;
{ : (;)  } [[ }' ;


: tDOES>
 \ Patch call to DOVAR to call end of parameter field instead:
  $  tCFA @ 5 +  -    \ Calculate jump distance
  tCFA @ 1+ T>H  dc!  \ Calculate location to patch; do it
 \ Compile this code at end of parameter field:
  [ { dodoes } ] LITERAL # call,  \ dodoes
  eax pop,                        \ get PFA
 \ Start compiling DOES> code:
  (:) ;


{ : ;    ;tCODE  } [[ }' ;
: :  ( "name" -) ( -)
  tCODE: ( add word to dictionaries)
  (:) ;


:' T: : ;
{ :' ;T ; }' ;




\ TESTS--------
\  hex $.
\  1234 tCON ABCDE
\  tCREATE ABCDF  5577 dc,
\  tVAR ABCDG
\  1122 tVAL ABCDH
\  tCREATE ABCDI  45434241 dc, tDOES>  ;

  \ Test meta-compiling
\  $.
\  (:) ABCDE 7978
\
\    ABCDF ABCDG ABCDH
\  (;)

  \ Test interpreting
\  { abcde . cr
\
\    abcdf . cr
\  }

\ save-buffers bye
