\ prim.f - Forth primitives                     21 Mar 2000  tcn

\ use string instructions where possible!

\ Register usage: (tentative)
\ EAX, EDX, ESI, EDI - Caller saves (Forth VM regs, used alot)
\ EBX, ECX - Callee saves (these are "extra" regs, used less)
\ ESP, EBP - Reserved (return & data stack, respectively)

DECIMAL ."    prim.f begins at " $.

HEX USE32



\ Constants
-1 tCON true
0 tCON false
tW tCON w


\ Address register:
\ TBD: APUSH, APOP to push/pop address reg before using it
\
\ TCODE: A    (  -a  (A -> TOS)
\   ESI DPUSH,  ;TCODE
\ TCODE: A!   ( a-   (TOS -> A)
\   ESI DPOP,  ;TCODE
\ TCODE: @A   (  -x  (Fetch from [A])
\   0PUSH,  0 ESI@ EAX MOV, ;TCODE
\ TCODE: @A+  (  -x  (Fetch from [A] then increment A)
\   0PUSH,  D-PTR LODS, ;TCODE
\ TCODE: C@A+ (  -x  (Fetch byte from [A] then increment A)
\   0PUSH,  LODS, ;TCODE
\ TCODE: !A   ( x-   (Store to [A])
\   EAX 0 ESI@ MOV, DROP, ;TCODE
\ TCODE: !A+  ( x-   (Store to [A] then increment A)
\   EAX 0 ESI@ MOV, 4 # ESI ADD, DROP, ;TCODE
\ TCODE: C!A+ ( x-   (Store byte to [A] then increment A)
\   EAX 0 ESI@ MOV, ESI INC, DROP, ;TCODE

\ Return Stack
{ :' push ( x-  )  EAX PUSH, DROP, }' ;
{ :' pop  (  -x )  0PUSH, EAX POP, }' ;
{ :' r@   (  -x )  0PUSH, 0 ESP@ EAX MOV, }' ;
{ :' rdrop         4 # ESP ADD, }' ;

\ Stack manipulation
TCODE: dup ( x - x x)
  0PUSH,  ;TCODE
TCODE: ?dup ( x - x x)  EAX EAX TEST,  0=, NOT, IF,
  0PUSH,  THEN,  ;TCODE
TCODE: swap ( x y - y x)
  EAX 0 EBP@ XCHG,  ;TCODE
TCODE: over ( x y - x y x)
  0PUSH,  tW EBP@ EAX MOV,  ;TCODE
TCODE: drop ( x -)  DROP,  ;TCODE
TCODE: 2drop ( x x -)
  tW EBP@ EAX MOV,  tW 2* # EBP ADD,  ;TCODE
TCODE: nip ( x y - y)  tW # EBP ADD,  ;TCODE

: 2dup  OVER OVER ;


\ TBD:
TCODE: rot  ( x1 x2 x3 - x2 x3 x1)  ebx push,
  0 ebp@ ebx mov, eax 0 ebp@ mov,
  tW ebp@ eax mov, ebx tW ebp@ mov,  ebx pop, ;TCODE
TCODE: -rot ( x1 x2 x3 - x3 x1 x2)  ebx push,
  tW ebp@ ebx mov, eax tW ebp@ mov,
  0 ebp@ eax mov, ebx 0 ebp@ mov,  ebx pop, ;TCODE

\ Memory access
TCODE: @  ( a - n)  0 EAX@ EAX MOV,         ;TCODE
TCODE: c@ ( a - c)  0 EAX@ EAX MOVZX,       ;TCODE
TCODE: 2@ ( a - n)  W-PTR 0 EAX@ EAX MOVZX, ;TCODE

TCODE: !  ( n a)
  esi push, 0 EBP@ ESI MOV,  ESI 0 EAX@ MOV,
  4 EBP@ EAX MOV,  tW 2* # EBP ADD,  esi pop, ;TCODE
TCODE: c! ( c a)
  ebx push,  0 EBP@ EBX MOV,  BL 0 EAX@ MOV,
  4 EBP@ EAX MOV,  tW 2* # EBP ADD,  ebx pop, ;TCODE
TCODE: 2! ( n a)
  ebx push,  0 EBP@ EBX MOV,  BX 0 EAX@ MOV,
  4 EBP@ EAX MOV,  tW 2* # EBP ADD,  ebx pop, ;TCODE

TCODE: +! ( n a)
  esi push, 0 EBP@ ESI MOV,  ESI 0 EAX@ ADD,
  4 EBP@ EAX MOV,  tW 2* # EBP ADD, esi pop, ;TCODE

tCODE: c@+ ( a - a+1 #)  ebx push,
  0 eax@ ebx movzx, eax inc, ebx dpush, ebx pop, ;tCODE
: count C@+ ;


tvar h
: here h @ ;
TCODE: , ( n)  edi push, { H } #@ EDI MOV,
  D-PTR STOS,  EDI { H } #@ MOV,  DROP,  edi pop, ;TCODE
TCODE: c, ( n)  edi push, { H } #@ EDI MOV,
        STOS,  EDI { H } #@ MOV,  DROP,  edi pop, ;TCODE
TCODE: 2, ( n)  edi push, { H } #@ EDI MOV,
  W-PTR STOS,  EDI { H } #@ MOV,  DROP,  edi pop, ;TCODE

TCODE: execute ( xt -)  ebx dpop, ebx call,  ;TCODE




\ Arithmetic
TCODE: + ( n n - n)  0 EBP@ EAX ADD,  tW # EBP ADD,  ;TCODE
TCODE: - ( n n - n)
  0 EBP@ EAX SUB, EAX NEG, tW # EBP ADD,  ;TCODE
TCODE: * ( n n - n)  edx push,
  d-ptr 0 EBP@ MUL,  tW # EBP ADD, edx pop, ;TCODE
TCODE: u/ ( n n - n)  edx push, 
  0 ebp@ eax xchg, CWD, d-ptr 0 EBP@ DIV,
  tW # EBP ADD, edx pop, ;TCODE
TCODE: umod ( n n - n)  edx push,
  0 ebp@ eax xchg, CWD, d-ptr 0 EBP@ DIV,
  EDX EAX MOV,  tW # EBP ADD, edx pop, ;TCODE
TCODE: u/mod ( u u - r q)  edx push,
  0 ebp@ eax xchg, cwd, d-ptr 0 ebp@ div,
  EDX 0 EBP@ MOV, edx pop, ;TCODE
TCODE: 1+ ( n - n+1)  EAX INC, ;TCODE
TCODE: 1- ( n - n-1)  EAX DEC, ;TCODE
TCODE: 2* ( n - 2n)   1 # EAX SHL, ;TCODE
TCODE: 2/ ( u - u/2)  1 # EAX SHR, ;TCODE
TCODE: w+ ( n - n)  4 # EAX ADD, ;TCODE
tCODE: w* ( n - n)  2 # eax shl, ;tCODE
TCODE: abs ( n - u)
  eax eax test,  0<, IF, eax neg, THEN, ;TCODE
TCODE: min ( n n - n)  ebx push, ebx dpop,
  ebx eax cmp, >, IF, ebx eax mov, THEN, ebx pop, ;tCODE
TCODE: max ( n n - n)  ebx push, ebx dpop,
  ebx eax cmp, <, IF, ebx eax mov, THEN, ebx pop, ;tCODE

\ TBD:
\ TCODE: um/mod ( l h u - r q  (Like U/MOD w/ 64-bit dividend)


\ Logic
TCODE: not ( n - n)  EAX COM,  ;TCODE
TCODE: neg ( n - n)  EAX NEG,  ;TCODE
TCODE: and ( n n - n)  0 EBP@ EAX AND,  tW # EBP ADD,  ;TCODE
TCODE: or  ( n n - n)  0 EBP@ EAX OR,   tW # EBP ADD,  ;TCODE
TCODE: xor ( n n - n)  0 EBP@ EAX XOR,  tW # EBP ADD,  ;TCODE
TCODE: lshift ( n # - n)  ecx push, EAX ECX MOV,
  0 EBP@ EAX MOV, EAX SHL, tW # EBP ADD, ecx pop, ;TCODE
TCODE: rshift ( n # - n)  ecx push, EAX ECX MOV,
  0 EBP@ EAX MOV, EAX SHR, tW # EBP ADD, ecx pop, ;TCODE
\ alias some ANS words (too long for my taste)
\ : INVERT NOT ;
\ : NEGATE NEG ;


\ Comparison operators

\ Most/all of these can be done without conditionals:
tCODE: 0= ( n - ?)  1 # eax sub,  eax eax sbb, ;tCODE
tCODE: 0<> ( n - ?)  1 # eax sub, eax eax sbb,  eax com, ;tCODE
tCODE: 0< ( n - ?)  32 # eax sar, ;tCODE
tCODE: 0> ( n - ?)  eax neg, 32 # eax sar, ;tCODE
tCODE: = ( n n - ?)
   0 ebp@ eax sub, ( -)
   1 # eax sub,  eax eax sbb, ( 0=)
   tW # ebp add, ( NIP) ;tCODE
tCODE: <> ( n n - ?)
   0 ebp@ eax sub, ( -)
   1 # eax sub,  eax eax sbb,  eax com, ( 0<>)
   tW # ebp add, ( NIP) ;tCODE

\ Untested optimizations:
\ : < - 0< ;
tCODE: < ( n n - ?)
   0 ebp@ eax sub, eax neg, ( -)
   32 # eax sar, ( 0<)  tW # ebp add, ( NIP) ;tCODE

\ : > - 0> ;
tCODE: > ( n n - ?)
   0 ebp@ eax sub, ( SWAP -)
   32 # eax sar, ( 0<)  tW # ebp add, ( NIP) ;tCODE

\ Untested:
tCODE: u< ( u u - ?)
   eax 0 ebp@ sub, eax eax sbb,
   tW # ebp add, ( NIP) ;tCODE


\ TBD: optimize the rest of them to avoid conditionals:
:' cond1,  ( unary)
  if, EAX EAX XOR,  EAX COM,
  else, EAX EAX XOR, then, ;
:' cond2,  ( binary)
  cond1, tW # EBP ADD, ;
 : <= ( n n - ?)  [[ EAX 0 EBP@ CMP,  >,  NOT,  cond2, ]] ;
 : >= ( n n - ?)  [[ EAX 0 EBP@ CMP,  <,  NOT,  cond2, ]] ;

\ TBD: cutting corners here, finish later!
: u> > ;
\ : u< < ;
: u>= >= ;


\ Misc (not classified yet)
: aligned ( a - a')
[[ 3 # al test,
   0=, not, if, 3 INVERT # al and, 4 # al add, then,  ]] ;
: align
  H @ ALIGNED H ! ;
: allot ( n)  H +! ;

tCODE: +under ( a b c - a+c b)
  edx push,  edx dpop, edx 0 ebp@ add,  edx pop, ;tCODE

{ :' exit  RET, }' ;


\ Looping
{ :' FOR ( n - )  ecx push,  ecx dpop,  BEGIN, }' ;
{ :' NEXT ( - )   LOOP, ecx pop, }' ;

{ :' BEGIN ( - ) }'  BEGIN, ;
{ :' AGAIN ( - ) }'  AGAIN, ;
{ :' UNTIL ( ? - ) }'  ebx push, ebx dpop,
  ebx ebx test, ebx pop,  0=, NOT, UNTIL, ;
{ :' WHILE ( ? - ) }'  ebx push, ebx dpop,
  ebx ebx or,   ebx pop,  0=, NOT, WHILE, ;
{ :' REPEAT ( - ) }' REPEAT, ;

\ Conditionals
{ :' IF ( ? - ) }'  ebx push,
  ebx dpop, ebx ebx or, ebx pop,  0=, not, if, ;
{ :' ELSE }'  else, ;
{ :' THEN }'  then, ;


\ Timing
: ms ( # -  (Millisecond delay) [[
  ecx push,
  6 # EAX SHL, ( multiply by 64, which is approx. 1000/15.085)
  EAX ECX XCHG, ( save ECX)
  61 # AL IN,  10 # AL AND,  AL AH MOV,
  BEGIN,
    BEGIN, 61 # AL IN,  10 # AL AND,  AH AL CMP, 0=, NOT, UNTIL,
    AL AH MOV,  LOOP,
  EAX ECX MOV,  DROP,  ecx pop, ]] ;

{ :' PAUSE  ( dummy) }' ;


\ Reboot
tCODE: boot
  cli, $1234 # $472 #@ mov,   \ Set BIOS "warm boot" flag
  $8D # al mov, $70 # al out, \ Disable NMI; Disable CMOS writes
  $71 # al in,
  esi esi xor, 0 # W-PTR esi #@ mov, \ Makes ints triple-fault
  $0F db, $01 db, $1E db, ( esi #@ lidt )
  int3, \ should triple-fault
  hang \ if that fails

