\ edit.f - block editor				11/3/99 tcn

1 CONSTANT BLOCK1	\ First block (0 or 1; varies)

: L>A ( line# - a) 6 LSHIFT ;      : A>L ( a - line#) 6 RSHIFT ;


VARIABLE DONE	\ set TRUE to exit editor
VARIABLE CUR	\ current block
VARIABLE POS	\ cursor position within block (linear, not x/y)
: >ADDR ( pos - addr) CUR @ BLOCK + ;
VARIABLE CUR'  VARIABLE POS'  \ Alternate (for 'last block' key)
1 CUR' !  0 POS' !

CREATE SCRATCH 1024 ALLOT  \ Cut/paste buffer

: UPDATE ( fix gForth's lousy UPDATE) CUR @ BLOCK DROP UPDATE ;

: BEEP 7 EMIT ;

\ What to print in place of unprintable characters:
254 CONSTANT UNPRINTABLE

: FILTER-EMIT ( c -) DUP 32 255 WITHIN
  0= IF DROP UNPRINTABLE  THEN  EMIT ;

( Print a 0-padded number)
: zeroes  dup IF 0 do [char] 0 emit loop ELSE drop THEN ;
: 0ud.r	>r <# #s #> r> over - zeroes type ;
: 0u.r	0 swap 0ud.r ;

: SHOW-CURRENT-CHAR
  BASE @   POS @ >ADDR C@
    66 4 AT-XY  10 BASE !  DUP 3 U.R  SPACE
    		." $" 16 BASE !  DUP 2 U.R  SPACE
    		." &" 8 BASE !   DUP 3 U.R  SPACE
    66 5 AT-XY	." %" 2 BASE !   8 0U.R
  BASE ! ;

: UPDATE-CURSOR ( -) POS @ 64 /MOD 2DUP
  66 2 AT-XY  SWAP  1+ 2 U.R ."  - " 1+ U.
  SHOW-CURRENT-CHAR    AT-XY ;

: UPDATE-STATUS
  66 0 AT-XY
\  CUR @ UPDATED? IF [CHAR] * ELSE BL THEN EMIT
  ."  #" CUR @ 1- U.
  UPDATE-CURSOR ( must be last) ;

: REPAINT
  PAGE
  CUR @ BLOCK
  15 FOR
    63 FOR  DUP C@ FILTER-EMIT CHAR+  NEXT
    [CHAR] | EMIT  CR  ( line on right)
  NEXT  DROP
  [CHAR] - 63 FOR DUP EMIT NEXT DROP  [CHAR] + EMIT  CR
  ( line on bottom)
  UPDATE-STATUS ;

: kLEFT	POS @ IF -1 POS +! UPDATE-CURSOR THEN ;
: kUP	POS @ 64 >= IF -64 POS +! UPDATE-CURSOR THEN ;
: kRIGHT  POS @ 1023 < IF 1 POS +! UPDATE-CURSOR THEN ;
: kDOWN	POS @ 960 < IF 64 POS +! UPDATE-CURSOR THEN ;
: kHOME	POS @ 960 AND POS ! UPDATE-CURSOR ;
: kEND	POS @ 63 OR POS ! UPDATE-CURSOR ; \ Go to last word!!
: kNEXT-WORD ;
: kPREV-WORD ;
: kNEXT-PAGE 1 CUR +! REPAINT ;
: kPREV-PAGE CUR @ BLOCK1 > IF -1 CUR +! THEN REPAINT ;

: kGOTO-BLOCK 0 17 AT-XY ." Go to block: " 0 0 PAD DUP 10 ACCEPT
  DUP IF >NUMBER 2DROP DROP ( block # typed) 1+ CUR !
  ELSE ( aborted) 2DROP 2DROP THEN  REPAINT ;

: kVIEW-LAST
  CUR @  CUR' @  CUR !  CUR' !
  POS @  POS' @  POS !  POS' !  REPAINT ;

: kENTER  POS @ 959 < IF kDOWN kHOME THEN ;

: DELETE-CHAR
  POS @ DUP 1+ SWAP ( pos+1 pos)
  DUP DUP 63 OR  SWAP - ( count)
  ROT >ADDR  ROT >ADDR  ROT  CMOVE
  BL  POS @ 63 OR  >ADDR  C! ( space at end)
  UPDATE REPAINT ;

: kDELETE  DELETE-CHAR ;
: kBACKSPACE  -1 POS +! DELETE-CHAR ;

: kDELETE-LINE
  POS @ 960 AND DUP >ADDR  DUP 64 + SWAP ROT ( from to tmp)
  960 SWAP - ( count) CMOVE
  960 >ADDR 64 BL FILL ( clear last line)  UPDATE REPAINT ;

: kCUT-LINE ;  \ TBD: Delete & copy to line-stack

: kYANK-LINE ; \ TBD: Insert line from line-stack

: kTOGGLE-OVERTYPE ;  \ TBD:

: INSERT ( c - c)
\ TBD: insert/overtype (currently just overtype)
  DUP 32 255 WITHIN  ( not a control char)
  POS @ 1023 <  ( not past end of block)
  AND IF
    DUP  POS @ >ADDR  C!  UPDATE ( store it)
    DUP EMIT  ( show it)  1 POS +! UPDATE-CURSOR
  THEN ;

\ Keys to avoid in gForth: ^C ^O ^Q ^S ^Z
\ (there must be a way to pass them through...?)
\ To handle keypad/function keys, we'll need to decode escape
\ codes (but it'll be easier in Retro, so I won't bother)

: EDITOR
  FALSE DONE !
  REPAINT
  BEGIN
    KEY CASE
      18 of	REPAINT  endof		\ ^R repaint
      24 of	TRUE DONE !  endof	\ ^X quit
      4 of	kDELETE  endof		\ ^D delete
      127 of	kDELETE  endof		\ DEL
      8 of	kBACKSPACE  endof	\ BS
      25 of	kDELETE-LINE  endof	\ ^Y delete line
\      25 of	kYANK-LINE  endof	\ ^Y yank (paste)
\ TBD: CUT LINE (save on line-stack)	^W ? (emacs keys)
\ TBD: PASTE LINE (from line-stack)  ^Y
\      9 of	kTAB  endof		\ TAB
      13 of	kENTER  endof		\ ENTER
      10 of	kLEFT  endof		\ ^J left
      11 of	kDOWN  endof		\ ^K down
      12 of	kRIGHT  endof		\ ^L right
      9 of	kUP  endof		\ ^I up
      1 of	kHOME  endof		\ ^A home (BOL)
      5 of	kEND  endof		\ ^E end (EOL)
      14 of	kNEXT-PAGE  endof	\ ^N next page
      2 of	kPREV-PAGE  endof	\ ^B back a page
      7 of	kGOTO-BLOCK  endof	\ ^G goto block
      22 of	kVIEW-LAST  endof	\ ^V view last block
      INSERT
    ENDCASE
  DONE @ UNTIL
  0 17 AT-XY ;



: PYGMY
  s" ~/forth/pygmy/pygmy.scr" OPEN-BLOCKS
  118 CUR !
  0 POS !
  EDITOR ;

: CMFORTH
  s" ~/forth/CMFORTH.BLK" OPEN-BLOCKS
  2 CUR !
  0 POS !
  EDITOR ;

: RETRO
  s" ~/fos/retro.fb" OPEN-BLOCKS
  1 CUR !
  0 POS !
  EDITOR ;

