\ web.f - Web browser stuff (in the prototype stage)

\ An idea:
\ Strings are dumped into PARSE-HTML as they are read, and PARSE-HTML
\ appends them to an output buffer, which can be read at any time while
\ the page is loading.

\ The output buffer could be in "chunks", for example,
\ COUNT/data/next --> COUNT/data/next --> COUNT/data/0 (end)
\
\ Output object format: 
\   4    next - ptr to next object 
\   4    type - ptr to display routine 
\   4,4  x,y  - position within document (rows/cols in text, pixels in graphics)
\   ...  data - type-specific data
\
\   Some data types:
\     Text: color, font, height, length, text...
\     Image: ptr to bitmap
\     HR: width, thickness
\
\   The point of this is to convert an ugly embedded-tag format (html) into
\   a structured format, like a CAD program.  For this purpose we can make
\   a simplification: we have a linked list in top-to-bottom order, for
\   fast displays.  We just index a few of the objects (one every page or
\   so, plus all the "anchors") so we can jump around the document quickly.
\
\   (Side note: If I ever get around to it, I'll use a similar method to
\   write a program that unifies word processing, raster/vector graphics,
\   maybe, spreadsheets, anything else that fits.  No reason to separate
\   things that are basically the same.)

only forth

\ Variables:
variable WS ( whitespace flag)
WS off


\ TAGS:
vocabulary HTML  also HTML definitions
: P  cr cr  WS off ;
: BR cr     WS off ;

get-current
only forth definitions
constant HTML-wordlist
: search-html ( a # - 0 | xt 1 )
  HTML-wordlist search-wordlist ;


\ ESCAPES: (i.e., "&amp;")
vocabulary ESCAPES  also ESCAPES definitions
: amp  '& emit ;
: lt   '< emit ;
: gt   '> emit ;
\ TBD: accents, umlauts, etc.

get-current
only forth definitions
constant ESC-wordlist
: search-esc ( a # - 0 | xt 1 )
  ESC-wordlist search-wordlist ;


\ Parsing HTML:
\   '<' - Parse an HTML tag delimited by ' ' or '>' and execute command
\         (in HTMLTAGS vocabulary)
\   '&' - Parse a symbol name delimited by ';' and output the symbol
\   ' ' - Output 1 space for any amount of whitespace (space/tab/cr/lf)
\   '.' - Output '. ' so we have 2 spaces after each period.

: getc ( a # - a' #' c)
  1- >r dup 1+ swap c@ r> swap ;

: IsSpace? ( c - f)
  dup 9 =  over 10 = or  over 13 = or  swap 32 = or ;

\ --------------------------
variable ST   \ State
: ST! ( # -)
\  ." ST" dup .
  ST ! ;

variable 0word \ Start current word

\ Enumerate states (they must be in the same order in JTBL)
0 constant #TXT
1 constant #WS
2 constant #ESC
3 constant #TAG
4 constant #PAR
5 constant #VAL

: >0word  dup 1+ 0word ! ;

\ Code for handling a character, for each state:
: stTXT ( c -)
  dup IsSpace? IF drop space #WS st! ELSE
    dup '< = IF drop #TAG st! >0word ELSE
      dup '& = IF drop #ESC st! >0word ELSE
        emit
      THEN
    THEN
  THEN ;

: stWS ( c -)
  dup IsSpace? invert IF #TXT st! stTXT ELSE drop THEN ;

: stESC ( c -)
  '; = IF
    0word @ 2dup - search-esc IF execute THEN
    #WS st!
  THEN ;

: stTAG ( c -)
  '> = IF
    0word @ 2dup - search-html IF execute THEN
    #WS st!
  THEN ;

: stPAR ( c -)
;

: stVAL ( c -)
;

create JTBL
' stTXT ,  ' stWS ,  ' stESC ,  ' stTAG ,  ' stPAR ,  ' stVAL ,

: doState ( c -  Execute the current state's routine)
  JTBL st @ cells + @ execute ;

: parse-html ( a # -  Feed me HTML source line-by-line)
  0 DO dup c@ doState 1+ LOOP drop ;

\ Initialize:
#WS st !

\ test:
s" asdf<p> this is       a paragraph<br>line breaks<br>&lt;&amp;escapes&gt;"
parse-html cr cr
